module PP where
import Text.PrettyPrint
import Syntax
import HtmlUtils
--import Utils(mix)
import Data.Char(toLower)
import Debug.Trace(trace)

{- This is a pretty printer that generates HTML from the abstract syntax.

   The first argument (es) to these function is a pair

	   (this,links) :: (String,[(String,String)]),

   where `this' is the URL of the web document being generated and links is
   a lookup table giving URLs to web page containing the definition of a
   name (used in prName). If you don't want any hyperlinks you can
   pass ("",[]),
-}

undef n x = trace (n++" documented but undefined.") x

ppValueDef es (n,Nothing) = undef n (text (n++" :: ??"))
ppValueDef es (_,Just d) = (ppSig es d) -- ok

ppTypeDef es (n,Nothing) = undef n (hsep [kw "type",text n,text "= ??"])
ppTypeDef es (_,Just d) =
  case d of
    Data d (Just cs) ->
        hsep [kw "data",ppCon es ppVar d]<+>
              sep (zipWith (<+>) (text "=":repeat (text "|")) (map ppCCon cs))
      where
	ppCCon (ctx,constr) = ppOptCtx es ctx <+> ppConstr constr
	ppConstr constr =
	  case constr of
	    PlainCon t ->  ppCon ("",[]) (ppAType es) t
	    LabelledCon con lbls ->
	        hsep [ppName ("",[]) con,braces $ sep $ punctuate comma (map ppLbl lbls)]
	      where
		ppLbl (names,t) =
		   hsep (punctuate comma (map (ppName es) names)++[text "::",ppType es t])
    Class d ms ->
        hsep [kw "class",ppCon es ppVar d,kw "where"] $$
	nest 3 (ppSigs es ms)
    _ -> ppTopDecl es d

ppTopDecl es dec =
    case dec of
      Data d Nothing -> hsep [kw "data",ppC d]
      Data d _ -> hsep [kw "data",ppC d,text "= ..."]
      Type d t -> hsep [kw "type",ppC d,text "=",ppType es t]
      Class d _ -> hsep [kw "class",ppC d,kw "where",text "..."]
      Instance [] d -> hsep [kw "  instance",ppCon es (ppAType es) d]
      Instance c d -> hsep [kw "  instance",ppCt c,plain "=>",ppCon es (ppAType es) d]
      Value sig -> ppSig es sig
      Infix f i n ->
        hsep [kw ("infix"++map toLower (show f)),int i,ppName es n]
  where
   ppC = ppCon es ppVar
   ppCt = ppCtx es ppVar

ppSigs es = vmap (ppSig es)

ppSig es (Sig f t) = ppName es f <+> plain "::" <+> ppCType es t

ppName es arg = ppName' es arg arg

ppName' (this,es) real arg = 
  case lookup real es of
    Just f | f/=this -> link (html f) (text arg)
    _ -> text arg -- on ("A NAME="++html arg)++escape arg++off "A"

ppVar v = wrap "VAR" (text v)

ppCType es (CT ctx t) = sep [ppOptCtx es ctx, ppType es t]

ppOptCtx es ctx =
    case ctx of
      [] -> empty
      _ -> ppCtx es ppVar ctx <+> plain "=>"

ppCtx es ppArg ctx =
  parens $ hsep $ punctuate comma (map (ppCon es ppArg) ctx)

ppType es t =
  case t of
    ConT (Fun t1 t2) -> sep [ppBType es t1 <+> plain "->", ppType es t2]
    _ -> ppBType es t

ppBType es t =
  case t of
    ConT ct@(Con _ _) -> ppCon es (ppAType es) ct
    AppT a ts -> hsep (ppVar a : map (ppAType es) ts)
    _ -> ppAType es t

ppAType es t =
  case t of
    VarT a -> ppVar a
    ConT ct@(List _) -> ppCon es (ppType es) ct
    ConT ct@(Tuple _) -> ppCon es (ppType es) ct
    ConT ct@(Con _ []) -> ppCon es (ppType es) ct
    t -> parens (ppType es t)

ppCon es ppArg c =
  case c of
    Con "Either" [a1,a2] -> ppArg a1<+>ppName' es "Either" "⊕"<+>ppArg a2
    Con c as -> ppName es c<+>sep (map ppArg as)
    List a -> brackets (ppArg a)
    Fun a b -> ppArg a <> plain "->" <> ppArg b
    Tuple as -> parens $ hsep $ punctuate comma (map ppArg as)
